# -*- coding: utf-8 -*-
# _calculateTF.py
# Module providing the calculateTF function
# Copyright 2013 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the calculateTF() function
"""

from warnings import warn

import numpy as np
from scipy.signal import lti, ss2zpk

from ._constants import eps
from ._partitionABCD import partitionABCD
from ._utils import carray, minreal


def calculateTF(ABCD, k=1.):
    """Calculate the NTF and STF of a delta-sigma modulator.

    The calculation is performed for a given loop filter
    ABCD matrix, assuming a quantizer gain of ``k``.

    **Parameters:**

    ABCD : array_like,
        The ABCD matrix that describes the system.

    k : float or ndarray-like, optional
        The quantizer gains. If only one quantizer is present, it may be set
        to a float, corresponding to the quantizer gain. If multiple quantizers
        are present, a list should be used, with quantizer gains ordered
        according to the order in which the quantizer inputs apperar in the
        ``C`` and ``D`` submatrices. If not specified, a default of one quatizer
        with gain ``1.`` is assumed.

    **Returns:**

    (NTF, STF) : a tuple of two LTI objects (or of two lists of LTI objects).

    If the system has multiple quantizers, multiple STFs and NTFs will be
    returned.

    In that case:

    * ``STF[i]`` is the STF from ``u`` to output number ``i``.
    * ``NTF[i, j]`` is the NTF from the quantization noise of the quantizer
      number ``j`` to output number ``i``.

    **Note:**

    Setting ``k`` to a list is unsupported in the MATLAB code (last checked
    Nov. 2014).

    **Example:**

    Realize a fifth-order modulator with the cascade-of-resonators structure,
    feedback form. Calculate the ABCD matrix of the loop filter and verify
    that the NTF and STF are correct.

    .. code-block:: python

        from deltasigma import *
        H = synthesizeNTF(5, 32, 1)
        a, g, b, c = realizeNTF(H)
        ABCD = stuffABCD(a,g,b,c)
        ntf, stf = calculateTF(ABCD)

    From which we get:

    ``H``::

             (z -1) (z^2 -1.997z +1) (z^2 -1.992z +0.9999)
        --------------------------------------------------------
         (z -0.7778) (z^2 -1.796z +0.8549) (z^2 -1.613z +0.665)

    coefficients::

        a: 0.0007, 0.0084, 0.055, 0.2443, 0.5579
        g: 0.0028, 0.0079
        b: 0.0007, 0.0084, 0.055, 0.2443, 0.5579, 1.0
        c: 1.0, 1.0, 1.0, 1.0, 1.0

    ABCD matrix::

        [[  1.00000000e+00   0.00000000e+00   0.00000000e+00   0.00000000e+00
            0.00000000e+00   6.75559806e-04  -6.75559806e-04]
         [  1.00000000e+00   1.00000000e+00  -2.79396240e-03   0.00000000e+00
            0.00000000e+00   8.37752565e-03  -8.37752565e-03]
         [  1.00000000e+00   1.00000000e+00   9.97206038e-01   0.00000000e+00
            0.00000000e+00   6.33294166e-02  -6.33294166e-02]
         [  0.00000000e+00   0.00000000e+00   1.00000000e+00   1.00000000e+00
           -7.90937431e-03   2.44344030e-01  -2.44344030e-01]
         [  0.00000000e+00   0.00000000e+00   1.00000000e+00   1.00000000e+00
            9.92090626e-01   8.02273699e-01  -8.02273699e-01]
         [  0.00000000e+00   0.00000000e+00   0.00000000e+00   0.00000000e+00
            1.00000000e+00   1.00000000e+00   0.00000000e+00]]

    NTF::

             (z -1) (z^2 -1.997z +1) (z^2 -1.992z +0.9999)
        --------------------------------------------------------
         (z -0.7778) (z^2 -1.796z +0.8549) (z^2 -1.613z +0.665)

    STF::

        1

    """

    nq = len(k) if type(k) in (tuple, list) else 1
    A, B, C, D = partitionABCD(ABCD, m=nq+1, r=nq)
    k = carray(k)
    diagk = np.atleast_2d(np.diag(k))

    B1 = B[:, 0]
    B2 = B[:, 1:]
    B1 = B1.reshape((B1.shape[0], 1)) if len(B1.shape) == 1 else B1
    B2 = B2.reshape((B2.shape[0], 1)) if len(B2.shape) == 1 else B2

    # In a single-quantizer system, D2 should be all zeros, as any
    # non-zero term in D2 would imply we got a delay-free loop.
    # The original MATLAB code implies so.
    # The trouble arises when we adapt and extend the code to calculate
    # the transfer functions for multiple-quantizer systems. There the
    # off-diagonal terms of D2 may very well be non-zero, therefore
    # in the following we consider D2.
    # this means that if you supply an ABCD matrix, with one quantizer
    # and an (erroneously) zero-delay loop, the MATLAB toolbox will
    # disregard D2 and pretend it's zero and the loop is correctly
    # delayed, spitting out the corresponding TF.
    # Instead here we print out a warning and process the ABCD matrix
    # with the user-supplied, non-zero D2 matrix.
    # The resulting TFs will obviously be different.

    D1 = D[:, 0]
    D2 = D[:, 1:]
    D1 = D1.reshape((D1.shape[0], 1)) if len(D1.shape) == 1 else D1
    D2 = D2.reshape((D2.shape[0], 1)) if len(D2.shape) == 1 else D2

    # WARN DELAY FREE LOOPS
    if np.diag(D2).any():
        warn("Delay free loop detected! D2 diag: %s", str(np.diag(D2)))

    # Find the noise transfer function by forming the closed-loop
    # system (sys_cl) in state-space form.
    Ct = np.linalg.inv(np.eye(nq) - D2*diagk)
    Acl = A + np.dot(B2, np.dot(Ct, np.dot(diagk, C)))
    Bcl = np.hstack((B1 + np.dot(B2, np.dot(Ct, np.dot(diagk, D1))),
                     np.dot(B2, Ct)))
    Ccl = np.dot(Ct, np.dot(diagk, C))
    Dcl = np.dot(Ct, np.hstack((np.dot(diagk, D1), np.eye(nq))))
    tol = min(1e-3, max(1e-6, eps**(1/ABCD.shape[0])))
    ntfs = np.empty((nq, Dcl.shape[0]), dtype=np.object_)
    stfs = np.empty((Dcl.shape[0],), dtype=np.object_)

    # sweep the outputs 'cause scipy is silly but we love it anyhow.
    for i in range(Dcl.shape[0]):
        # input #0 is the signal
        # inputs #1,... are quantization noise
        stf_z, stf_p, stf_k  = ss2zpk(Acl, Bcl, Ccl[i, :], Dcl[i, :], input=0)
        stf = lti(stf_z, stf_p, stf_k)
        for j in range(nq):
            ntf_z, ntf_p, ntf_k = ss2zpk(Acl, Bcl, Ccl[i, :], Dcl[i, :], input=j+1)
            ntf = lti(ntf_z, ntf_p, ntf_k)
            stf_min, ntf_min = minreal((stf, ntf), tol)
            ntfs[i, j] = ntf_min
        stfs[i] = stf_min

    # if we have one stf and one ntf, then just return those in a list
    if ntfs.shape == (1, 1):
        return [ntfs[0, 0], stfs[0]]
    return ntfs, stfs




# -*- coding: utf-8 -*-
# test_synthesizeQNTF.py
# This module provides the tests for the synthesizeQNTF function.
# Copyright 2015 Giuseppe Venturini
# This file is distributed with python-deltasigma.
#
# python-deltasigma is a 1:1 Python port of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.
#

"""This module provides the test class for the synthesizeQNTF() function.
"""

from __future__ import division, print_function

import unittest

import numpy as np

from deltasigma import pretty_lti, synthesizeQNTF

class TestSynthesizeQNTF(unittest.TestCase):
    """Test class for synthesizeQNTF()"""

    def setUp(self):
        pass

    def test_synthesizeQNTF_1(self):
        """Test function for synthesizeQNTF() 1/4"""
        ntf = synthesizeQNTF(4, 32, 1/16, -50, -10)
        z, p, k = ntf
        p_ref = [0.573877782470855 + 0.569921695571522j,
                 0.808788367241398 + 0.002797375873482j,
                 0.591250299914031 + 0.244903892981552j,
                 0.673072277003855 - 0.278795665592338j]
        z_ref = [0.888037198535288 + 0.459771610712905j,
                 0.953044748762363 + 0.302829501298050j,
                 0.923879532511340 + 0.382683432365112j,
                 0.923879532511287 - 0.382683432365090j]
        k_ref = 1.
        np.allclose(p, p_ref, atol=1e-4, rtol=1e-3)
        np.allclose(z, z_ref, atol=1e-4, rtol=1e-3)
        np.allclose(k, k_ref, atol=1e-4, rtol=1e-3)

    def test_synthesizeQNTF_2(self):
        """Test function for synthesizeQNTF() 2/4"""
        ntf = synthesizeQNTF(4, 32, 1/16, -50, -10, 0)
        z, p, k = ntf
        z_ref = [0.910594282901269 + 0.413301405692653j,
                 0.936135618988396 + 0.351639165709982j,
                 0.888265620891369 + 0.459330150047295j,
                 0.952894107929043 + 0.303303180125290j]

        p_ref = [0.767590403998773 + 0.239841808290628j,
                 0.712362148895601 + 0.373174610786908j,
                 0.899297507221408 + 0.158408031000048j,
                 0.747910758574959 + 0.523887972745873j]
        k_ref = 1.
        np.allclose(p, p_ref, atol=1e-4, rtol=1e-3)
        np.allclose(z, z_ref, atol=1e-4, rtol=1e-3)
        np.allclose(k, k_ref, atol=1e-4, rtol=1e-3)

    def test_synthesizeQNTF_3(self):
        """Test function for synthesizeQNTF() 3/4"""
        order = 4
        osr = 32
        NG = -50
        ING = -10
        f0 = 1./16
        ntf0 = synthesizeQNTF(order, osr, f0, NG, ING)
        zeros_ref = np.array((0.8880 + 0.4598j,
                              0.9530 + 0.3028j,
                              0.9239 + 0.3827j,
                              0.9239 - 0.3827j))
        poles_ref = np.array((0.5739 + 0.5699j,
                              0.8088 + 0.0028j,
                              0.5913 + 0.2449j,
                              0.6731 - 0.2788j))
        # round down to the same precision as the results data
        # or np.sort_complex will find slightly different floats
        # to be different and sort wrongly, resulting in a failed test
        zeros = np.round(ntf0[0], 4)
        poles = np.round(ntf0[1], 4)
        assert np.allclose(np.sort_complex(zeros), np.sort_complex(zeros_ref),
                           atol=1e-2, rtol=1e-2)
        assert np.allclose(np.sort_complex(poles), np.sort_complex(poles_ref),
                           atol=1e-2, rtol=1e-2)
        assert ntf0[2] == 1.

    def test_synthesizeQNTF_4(self):
        """Test function for synthesizeQNTF() 4/4"""
        # order 2
        order = 2
        osr = 32
        NG = -50
        ING = -10
        f0 = 1 / 16
        ntf0 = synthesizeQNTF(order, osr, f0, NG, ING)
        z_ref = [0.900716472438935 + 0.434407454214544j,
                 0.944075182261086 + 0.329730268914908j]
        p_ref = [0.112333561599987 - 0.126178981177517j,
                 -0.00979019007164386 + 0.168653836396020j]
        k_ref = 1
        assert np.allclose(np.sort(z_ref), np.sort(ntf0[0]))
        assert np.allclose(np.sort(p_ref), np.sort(ntf0[1]))
        assert ntf0[2] == k_ref




# -*- coding: utf-8 -*-
# _simulateQSNR.py
# Module providing the simulateQSNR function
# Copyright 2015 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the simulateQSNR() function
"""

from __future__ import division, print_function

from warnings import warn

import numpy as np
from numpy.fft import fft, fftshift

from ._calculateSNR import calculateSNR
from ._simulateQDSM import simulateQDSM

def simulateQSNR(ntf,
                 R=64,
                 amp=None,
                 f0=0,
                 nlev=2,
                 f=None,
                 k=13):
    """
    Determine the SNR for a quadrature delta-sigma modulator using simulations.

    The modulator is described by a Noise Transfer Function (NTF)
    and the number of quantizer levels.

    Using sine waves located in FFT bins, the SNR is calculated as the ratio
    of the sine wave power to the power in all in-band bins other than those
    associated with the input tone. Due to spectral smearing, the input tone
    is not allowed to lie in bins 0 or 1.

    **Parameters:**

    ntf : tuple, ndarray or LTI object
        The Noise Transfer Function in any form supported by
        :func:`simulateQDSM`, such as an ABCD matrix or an NTF description,
        for example a zpk tuple, num-den tuple or an LTI object.
        If no information is available regarding the STF, it is assumed to
        be unitary.
    R : int, optional
        The oversampling ratio, defining the band of interest. Defaults to 64.
    amp : sequence, optional
        The sequence of the amplitudes to be used for the input signal, in
        ascending order, expressed in dB, where 0 dB means a full-scale (peak
        value :math:`n_{lev}-1`) sine wave. Defaults to [-120 -110...-20 -15 -10
        -9 -8 ... 0] dB.
    f0 : float, optional
        The normalized center frequency of the modulator. Defaults to 0.
    nlev : int, optional
        The number of levels in the modulator quantizer. Defaults to 2.
    f : float, optional
        The input signal frequency, normalized such that :math:`1 \\rightarrow
        f_s`. It is rounded to an FFT bin. If not set, defaults to
        :math:`1/(4\\cdot OSR)`.
    k : int, optional
        The integer ``k`` sets the length of the FFT, which is :math:`2^k`.
        Defaults to 13.

    **Returns:**

    snr : ndarray
        The calculated SNR.
    amp : ndarray
        The amplitude vector corresponding to the SNR.

    """
    if amp is None:
        amp = np.concatenate((np.arange(-120, -20+1, 10),
                              np.atleast_1d(-15),
                              np.arange(-10, 1)))
        if f is None:
            f = f0 + 1./(4*R)
    if np.abs(f - f0) > 1./(2 * R):
        warn('The input tone is out-of-band.')
    N = 2**k
    if N < 8*R:
        # Require at least 8 bins to be "in-band"
        warn('Increasing k to accommodate a large oversampling ratio.')
        k = np.ceil(np.log2(8*R))
        N = 2**k
    F = np.round(f*N)
    if F <= 1:
        warn('Increasing k to accommodate a low input frequency.')
        # We want f*N > 1
        k = np.ceil(np.log2(1./f))
        N = 2**k
        F = 2

    Ntransient = 100
    tone = (nlev - 1)*np.exp(2j*np.pi*F/N*np.arange(-Ntransient, N))
    # Hanning window of length N
    window = 0.5*(1 - np.cos(2*np.pi*np.arange(N)/N))
    f1 = max((np.round(N*(0.5 + f0 - 0.5/R)), 0))
    inBandBins = np.arange(f1, np.round(N*(0.5 + f0 + 0.5/R)) + 1, dtype=np.int32)
    F = F - f1 + N/2.

    snr = np.zeros(amp.shape)
    i = 0
    for A in 10.0**(amp/20.):
        v, _, _, _ = simulateQDSM(A*tone, ntf, nlev)
        hwfft = fftshift(fft(window*v[Ntransient:N + Ntransient]))
        snr[i] = calculateSNR(hwfft[inBandBins], F)
        i = i + 1
    return snr, amp



from distutils.core import setup
from distutils.extension import Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext

import numpy

extensions = [
    Extension("_simulateQDSM_core",
              ["_simulateQDSM_core.py",
               "_simulateQDSM_core.pxd"],
        include_dirs = [numpy.get_include()],
        libraries = [],
        library_dirs = []),
#    Extension("*", ["*.pyx"],
#        include_dirs = [numpy.get_include()],
#        libraries = [],
#        library_dirs = []),
]
setup(
    name = "My hello app",
    ext_modules = extensions,
    #ext_modules = cythonize(extensions),
    cmdclass = {'build_ext':build_ext}
)



# -*- coding: utf-8 -*-
# _synthesizeQNTF.py
# Module providing the synthesizeQNTF function
# Copyright 2013 Giuseppe Venturini
# This file is distributed with python-deltasigma.
#
# python-deltasigma is a 1:1 Python port of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.
#

"""
Module providing the Quadrature NTF synthesis function.
"""

from __future__ import division, print_function

from warnings import warn

import numpy as np

from scipy.signal import cheby2
from scipy.linalg import norm

from ._evalTF import evalTF
from ._dbv import dbv
from ._rmsGain import rmsGain
from ._synthesizeNTF import synthesizeNTF
from ._utils import _get_zpk

# Maximum number of iterations to reach the obj
ITN_MAX = 20

def synthesizeQNTF(order=4, OSR=64, f0=0., NG=-60, ING=-20, n_im=None):
    """Synthesize a noise transfer function for a quadrature modulator.

    **Parameters:**

    order : int, optional
        The order of the modulator. Defaults to 4.

    OSR : int, optional
        The oversampling ratio. Defaults to 64.

    f0 : float, optional
        The center frequency, normalized such that :math:`1 \\rightarrow f_s`.
        Defaults to 0, ie to a low-pass modultor.

    NG : float, optional
        The in-band noise gain, expressed in dB. Defaults to -60.

    ING : float, optional
        The image-band noise gain, in dB. Defaults to -20.

    n_im : int, optional
        The number of in-band image zeros, defaults to ``floor(order/3)``.

    **Returns:**

    ntf : (z, p, k) tuple
        ``ntf`` is a zpk tuple containing the zeros, poles and the gain of the
        synthesized NTF.

    .. note::

        From the Matlab Delta-Sigma Toobox:
        ALPHA VERSION:
        This function uses an experimental ad-hoc method that is
        neither optimal nor robust.

    **Example:**

    Fourth order quadrature modulator::

        from deltasigma import *
        order = 4
        osr = 32
        NG = -50
        ING = -10
        f0 = 1 / 16
        ntf0 = synthesizeQNTF(order, osr, f0, NG, ING)
        pretty_lti(ntf0)

    Returns::

          (z - 0.888 - 0.4598j) (z - 0.9239 + 0.3827j) (z - 0.9239 - 0.3827j) (z - 0.953 - 0.3028j)  
          ---------------------------------------------------------------------------------------------
           (z - 0.5739 - 0.5699j) (z - 0.5913 - 0.2449j) (z - 0.6731 + 0.2788j) (z - 0.8088 - 0.0028j) 


    .. image:: _static/synthesizeQNTF.png


    """
    if n_im is None:
        n_im = np.floor(order/3)
    debug_it = 0
    if n_im == 0:
        # Use synthesizeNTF to get an NTF with the specified NG; ignore ING
        f1 = 0.5/OSR
        x = 1.5
        lowest_f = np.inf
        dfdx = None
        for itn in range(ITN_MAX):
            ntf = synthesizeNTF(order, OSR, 1., x)
            f = dbv(rmsGain(ntf, 0., f1)) - NG
            if debug_it:
                print('x=%.2f f=%.2f' % (x, f))
            if abs(f) < 0.01:
                break
            if dfdx is None:
                dx = 0.1*np.sign(f)
                dfdx = 0
            else:
                dfdx = (f - f_old)/dx
                dx_old = dx
                dx = - f/dfdx
                if abs(dx) > max((1, 2*abs(dx_old))):
                    dx = np.sign(dx)*max((1, 2*abs(dx_old)))
                if x + dx <= 1:
                    # Hinf must be at least 1
                    dx = dx/2.
            f_old = f
            x = x + dx
        if itn == ITN_MAX - 1:
            warn('Iteration limit reached. NTF may be poor.')
        # Rotate the NTF
        z0 = np.exp(2j*np.pi*f0)
        zeros, poles, k = _get_zpk(ntf)
        ntf = (z0*zeros, z0*poles, k)
    else:
        n_in = order - n_im
        f1 = f0 - 0.5/OSR
        f2 = f0 + 0.5/OSR
        z0 = np.exp(2j*np.pi*f0)
        x = np.array([20., 20.])
        # "R" parameters for cheby2()
        lowest_f = np.inf
        dfdx = np.array([float('NaN'), float('NaN')])
        freq = np.linspace(-0.5, 0.5, 200)
        for itn in range(ITN_MAX):
            if debug_it:
                print('\nx = [%.2f, %.2f]' % (x[0], x[1]))
            b1, a1 = cheby2(n_in, x[0], 1./OSR, 'highpass')
            b2, a2 = cheby2(n_im, x[1], 1./OSR, 'highpass')
            ntf0 = (np.concatenate((np.roots(b1)*z0, np.roots(b2)*np.conj(z0))),
                    np.concatenate((np.roots(a1)*z0, np.roots(a2)*np.conj(z0))),
                    1)
            m = evalTF(ntf0, np.exp(2j*np.pi*freq))
            NG0 = dbv(rmsGain(ntf0, f1, f2))
            ING0 = dbv(rmsGain(ntf0, -f1, -f2))
            if debug_it:
                import pylab as plt
                from ._plotPZ import plotPZ
                from ._figureMagic import figureMagic
                plt.figure()
                plt.subplot(121)
                plotPZ(ntf0)
                plt.subplot(122)
                print('NG = %.1f, ING= %.1f' % (NG0, ING0))
                plt.plot(freq, dbv(m))
                figureMagic([-0.5, 0.5], 0.05, 2, [-100, 30], 10, 2)
                plt.hold(True)
                plt.plot([f1, f2], [NG0, NG0], 'k')
                plt.text(np.mean([f1, f2]), NG0, ('NG=%.1fdB' % NG0), va='bottom')
                plt.plot([-f1, -f2], [ING0, ING0], 'k')
                plt.text(np.mean([-f1, -f2]), ING0, ('ING=%.1fdB' % ING0), va='bottom')
                plt.show()
            f = np.array([NG0 - NG, ING0 - ING])
            if max(abs(f)) < 0.01:
                break
            if norm(f) < lowest_f:
                lowest_f = norm(f)
                # ntf0 is ALREADY a zpk tuple
                zeros, poles, k = _get_zpk(ntf0)
                best = (zeros.copy(), poles.copy(), k)
            if abs(f[0]) > abs(f[1]):
                # adjust x(1)
                i = 0
            else:
                # adjust x(2)
                i = 1
            if np.isnan(dfdx[i]):
                dx = np.sign(f[i])
                dfdx[i] = 0
                dfdx[1 - i] = float('NaN')
            else:
                dfdx[i] = (f[i] - f_old[i])/dx
                dfdx[1 - i] = float('NaN')
                dx = -f[i]/dfdx[i]
                xnew = x[i] + dx
                if xnew < 0.5*x[i]:
                    dx = -0.5*x[i]
                else:
                    if xnew > 2*x[i]:
                        dx = x[i]
            f_old = f.copy()
            x[i] = x[i] + dx
        if itn == ITN_MAX - 1:
            warn('Iteration limit reached. NTF may be poor')
        ntf = best
    return ntf




# -*- coding: utf-8 -*-
# test_simulateQSNR.py
# This module provides the tests for the simulateQSNR function.
# Copyright 2015 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's 
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""This module provides the test class for the simulateSNR() function.
"""

from __future__ import division, print_function

import unittest
import numpy as np
import deltasigma as ds

class TestSimulateQSNR(unittest.TestCase):
    """Test class for simulateQSNR()"""

    def setUp(self):
        self.SNR_ref = np.array([23.0421, 32.1100, 43.3758, 53.1791,
                                 65.5504, 70.5023, 73.4608, 76.2416, 77.8770,
                                 78.2733, 79.3729, 79.5728, 80.8729, 82.7461,
                                 83.0723, 84.8488, 84.3327])
        self.AMP_ref = np.array([-70, -60, -50, -40, -30, -20, -15, -10, -9, -8,
                                 -7, -6, -5, -4, -3, -2, -1, 0])


    def test_simulateQSNR(self):
        """Test function for simulateSNR() 4/4"""
        order = 4
        osr = 32
        M = 8
        NG = -50
        ING = -10
        f0 = 1./ 16
        quadrature = 1
        form = 'PFB'
        nlev = M + 1
        z0 = np.exp(1j*2*np.pi*f0)
        bw = 1./ osr
        delta = 2
        FullScale = M
        ntf0 = ds.synthesizeQNTF(order, osr, f0, NG, ING)
        ABCD = ds.realizeQNTF(ntf0, form, True)
        a, b = ds.simulateQSNR(ABCD, osr, None, f0, nlev);
        assert np.allclose(a[6:], self.SNR_ref, atol=6, rtol=1e-3)
        assert np.allclose(b[5:], self.AMP_ref, atol=1)



from __future__ import division

import numpy as np
import copy

def designLCBP6(n=3, OSR=64, opt=2, Hinf=1.6, f0=0.25, t=None, form='FB', x0=None, dbg=False):
    param = {'n':n, 'OSR':OSR, 'Hinf':Hinf, 'f0':f0, 't':t, 'form':form}
    if opt:
        w = 2*np.pi*(f0 + 0.25/OSR*ds_optzeros(n, opt))
    else:
        w = 2*np.pi*f0*np.ones(n)
    param.update({'l':1.0/w})
    param.update({'c':1.0/w})
    if x0 is None:
        x = np.arange(2, n+1)
    else:
        if len(x0) != n:
            raise ValueError('Initial guess (x0) has the wrong number of parameters.')
        else:
            x = copy(x0)
    if dbg:
        fprintf(1,char('Iterations for initial pole placement:\n'))
    options=optimset(char('TolX'),0.001,char('TolFun'),1,char('MaxIter'),1000)
    options=optimset(options,char('LargeScale'),char('off'))
    x=fmincon(LCObj1a,x,[],[],[],[],[],[],LCObj1b,options,param,0.97,dbg)
    H=LCoptparam2tf(x,param)
    rmax=max(abs_(H.p[:]))
    if rmax > 0.97:
        fprintf(2,char('Unable to find parameter values which stabilize the system.\n'))
        return param,H,L0,ABCD,x
    if dbg:
        fprintf(1,char('\nParameter Optimization:\n'))
    options=optimset(char('TolX'),0.0001,char('TolFun'),0.5,char('TolCon'),0.01)
    options=optimset(options,char('MaxIter'),1000)
    options=optimset(options,char('DiffMaxChange'),0.01)
    options=optimset(options,char('DiffMinChange'),0.0001)
    options=optimset(options,char('LargeScale'),char('off'))
    if dbg > 1:
        options=optimset(options,char('Display'),char('iter'))
    x=fmincon(LCObj2a,x,[],[],[],[],[],[],LCObj2b,options,param,0.97,dbg)
    H,L0,ABCD,param=LCoptparam2tf(x,param,nargout=4)
    gain_f0=abs_(evalTFP(L0,H,f0))
    param.gu = copy(param.gu / gain_f0)
    L0.k = copy(L0.k / gain_f0)
    ABCD[:,2 * n + 1]=ABCD[:,2 * n + 1] / gain_f0
    if strcmp(form,char('FF')) or strcmp(form,char('FFR')):
        param.gu[n + 1]=1
    if dbg:
        LCplotTF(H,L0,param)
    return param,H,L0,ABCD,x

def _LCObj1a(x=None, param=None, max_radius=None, dbg=None):
    """The objective function for the initial optimization
    process used to put the roots of the denominator of the LCBP NTF inside
    the unit circle."""
    return 1. # No actual objective

def LCObj1b(x=None,param=None,max_radius=None,dbg=None,*args,**kwargs):
    """The constraints function for the initial optimization
    process used to put the roots of the denominator of the LCBP NTF inside
    the unit circle."""
    H = LCoptparam2tf(x,param)
    objective = 1
    rmax=max(abs_(H.p[:]))
    C = rmax - max_radius
    Ceq = np.array([])
    if dbg:
        print('x = [ %.4f ]\n' % x)
        print('rmax = %f\n\n' % rmax)
    return C, Ceq

def LCObj2a(x=None,param=None,rmax=None,dbg=None,*args,**kwargs):
    nargout = kwargs["nargout"] if kwargs else None
    varargin = cellarray(args)
    nargin = 4-[x,param,rmax,dbg].count(None)+len(args)

    if dbg:
        fprintf(1,char('x = ['))
        fprintf(1,char('%8.5f '),x)
        fprintf(1,char(']\n'))
    H,L0,ABCD=LCoptparam2tf(x,param,nargout=3)
    f1=param.f0 - 0.25 / param.OSR
    f2=param.f0 + 0.25 / param.OSR
    objective=dbv(rmsGain(H,f1,f2) / sqrt(3 * param.OSR))
    fprintf(1,char('N0^2 = %.1fdB\n'),objective)
    return objective

def LCObj2b(x=None,param=None,rmax=None,dbg=None,*args,**kwargs):
    nargout = kwargs["nargout"] if kwargs else None
    varargin = cellarray(args)
    nargin = 4-[x,param,rmax,dbg].count(None)+len(args)

    H,L0,ABCD=LCoptparam2tf(x,param,nargout=3)
    Ceq=matlabarray([])
    max_radius=max(abs_(H.p[:]))
    H_inf=infnorm(H)
    stf0=abs_(evalTFP(L0,H,param.f0))
    tmp1,tmp2,tmp3,y=simulateDSM(0.5 / stf0 * sin(2 * pi * param.f0 * cat(arange(0,1000))),ABCD,nargout=4)
    ymax=max(abs_(y))
    if strcmp(param.form,char('FB')):
        f1=param.f0 - 0.25 / param.OSR
        f2=param.f0 + 0.25 / param.OSR
        stf1=abs_(evalTFP(L0,H,f1)) / stf0
        stf2=abs_(evalTFP(L0,H,f2)) / stf0
        C=matlabarray(cat(100 * (max_radius - rmax),H_inf - param.Hinf,stf1 - 1.01,stf2 - 1.01(ymax - 5) / 10))
    else:
        C=matlabarray(cat(100 * (max_radius - rmax),H_inf - param.Hinf(ymax - 5) / 10))
    if dbg:
        fprintf(1,char('constraints ='))
        fprintf(1,char(' %8.5f'),C)
        fprintf(1,char('\n'))
        fprintf(1,char('rmax=%5.3f, Hinf = %4.2f, ymax = %5.3f\n\n'),max_radius,H_inf,ymax)
    return C,Ceq



# -*- coding: utf-8 -*-
# _simulateQDSM.py
# Module providing the simulateQDSM function
# Copyright 2015 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the simulateQDSM() core function
"""

import numpy
cimport numpy as np
import cython

#@cython.wraparound(False)
#@cython.nonecheck(False)
#@cython.boundscheck(False)
@cython.locals(N=cython.int, k=cython.float, v=np.ndarray,
               y=np.ndarray,xn=np.ndarray,xmax=np.ndarray)
 
#def simulateQDSM_core(u, A, B, C, D1, order, nlev, nq, x0):
cpdef inline simulateQDSM_core(np.ndarray[complex, ndim=2] u,
                               np.ndarray[complex, ndim=2] A,
                               np.ndarray[complex, ndim=2] B,
                               np.ndarray[complex, ndim=2] C,
                               np.ndarray[complex, ndim=2] D1,
                               int order, nlev, int nq,
                               np.ndarray[complex, ndim=2] x0)

@cython.locals(v=np.ndarray, ytmp=np.ndarray)
cdef inline ds_qquantize(np.ndarray[complex, ndim=1] y, n)




# -*- coding: utf-8 -*-
# test_calculateTF.py
# This module provides the tests for the calculateTF() function.
# Copyright 2014 Giuseppe Venturini & Shayne Hodge
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

import unittest
import numpy as np
import deltasigma as ds
from deltasigma._utils import cplxpair


class TestCalculateTF(unittest.TestCase):
    """Test function for calculateTF()"""

    def setUp(self):
        ABCD = [[1.0, 0.0, 0.0, 0.044408783846879, -0.044408783846879],
                [0.999036450096481, 0.997109907515262, -0.005777399147297,
                 0.0, 0.499759089304780],
                [0.499759089304780, 0.999036450096481, 0.997109907515262,
                 0.0, -0.260002096136488],
                [0.0, 0.0, 1.0,  0.0, 0.0]]
        ABCD = np.array(ABCD)
        ntf, stf = ds.calculateTF(ABCD)
        ntf_zeros, ntf_poles, _ = ntf
        stf_zeros, stf_poles, _ = stf
        mntf_poles = np.array((1.498975311463384, 1.102565142679772,
                               0.132677264750882))
        mntf_zeros = np.array((0.997109907515262 + 0.075972576202904j,
                               0.997109907515262 - 0.075972576202904j,
                               1.000000000000000 + 0.000000000000000j))
        mstf_zeros = np.array((-0.999999999999996,))
        mstf_poles = np.array((1.498975311463384, 1.102565142679772,
                               0.132677264750882))

        # for some reason, sometimes the zeros are in different order.
        self.ntf_zeros, self.mntf_zeros = (cplxpair(ntf_zeros),
                                           cplxpair(mntf_zeros))
        self.stf_zeros, self.mstf_zeros = (cplxpair(stf_zeros),
                                           cplxpair(mstf_zeros))
        self.ntf_poles, self.mntf_poles = (cplxpair(ntf_poles),
                                           cplxpair(mntf_poles))
        self.stf_poles, self.mstf_poles = (cplxpair(stf_poles),
                                           cplxpair(mstf_poles))

    def test_calculateTF1(self):
        """Test function for calculateTF() 1/4"""
        # ntf zeros
        self.assertTrue(np.allclose(self.ntf_zeros, self.mntf_zeros, rtol=1e-5,
                        atol=1e-8))
        # ntf poles
        self.assertTrue(np.allclose(self.ntf_poles, self.mntf_poles, rtol=1e-5,
                        atol=1e-8))
        # stf zeros
        self.assertTrue(np.allclose(self.stf_zeros, self.mstf_zeros, rtol=1e-5,
                        atol=1e-8))
        # stf poles
        self.assertTrue(np.allclose(self.stf_poles, self.mstf_poles, rtol=1e-5,
                        atol=1e-8))

    def test_calculateTF2(self):
        """Test function for calculateTF() 2/4"""
        # test an easy TF
        ABCD = np.array([[1., 1., -1.],
                        [1., 0., 0.]])
        k = 1.
        ntf, stf = ds.calculateTF(ABCD, k) 
        ntf_zeros, ntf_poles, ntf_gain = ntf
        stf_zeros, stf_poles, stf_gain = stf
        self.assertTrue(np.allclose(stf_poles, [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(not len(stf_zeros))
        self.assertTrue(np.allclose(stf_gain, 1., rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntf_poles, [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntf_zeros, [1.], rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntf_gain, 1., rtol=1e-5, atol=1e-8))

    def test_calculateTF3(self):
        """Test function for calculateTF() 3/4"""
        # test for the default k value
        ABCD = np.array([[1., 1., -1.],
                        [1., 0., 0.]])
        ntf, stf = ds.calculateTF(ABCD)
        ntf_zeros, ntf_poles, ntf_gain = ntf
        stf_zeros, stf_poles, stf_gain = stf
        self.assertTrue(np.allclose(stf_poles, [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(not len(stf_zeros))
        self.assertTrue(np.allclose(stf_gain, 1., rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntf_poles, [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntf_zeros, [1.], rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntf_gain, 1., rtol=1e-5, atol=1e-8))

    def test_calculateTF4(self):
        """Test function for calculateTF() 4/4"""
        # Easy test for a 2-quantizers system
        # MASH 1-0 cascade
        ABCD = [[1, 1, -1, 0],
                [1, 0, 0, 0],
                [1, 0, -1, 0]]
        ABCD = np.array(ABCD, dtype=np.float_)
        k = [1., 1.]
        # here we get back arrays of transfer functions
        ntfs, stfs = ds.calculateTF(ABCD, k=k)
        # stfs
        self.assertTrue(np.allclose(stfs[0][1], [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(not len(stfs[0][0]))
        self.assertTrue(np.allclose(stfs[0][2], 1., rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(stfs[1][1], [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(not len(stfs[1][0]))
        self.assertTrue(np.allclose(stfs[1][2], 1., rtol=1e-5, atol=1e-8))
        # e1 to V1
        self.assertTrue(np.allclose(ntfs[0, 0][1], [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntfs[0, 0][0], [1.], rtol=1e-5, atol=1e-8))
        self.assertTrue(np.allclose(ntfs[0, 0][2], 1., rtol=1e-5, atol=1e-8))
        # e1 to V2
        self.assertTrue(np.allclose(ntfs[1, 0][1], [0.], rtol=1e-5, atol=1e-8))
        self.assertTrue(not len(ntfs[1, 0][0]))
        self.assertTrue(np.allclose(ntfs[1, 0][2], -1., rtol=1e-5, atol=1e-8))
        # e2 to V2
        self.assertTrue(not len(ntfs[1, 1][0]))
        self.assertTrue(not len(ntfs[1, 1][1]))
        self.assertTrue(np.allclose(ntfs[1, 1][2], 1., rtol=1e-5, atol=1e-8))
        # e2 to V1
        self.assertTrue(np.allclose(ntfs[0, 1][2], 0., rtol=1e-5, atol=1e-8))




# -*- coding: utf-8 -*-
# _realizeQNTF.py
# Module providing the realizeQNTF function
# Copyright 2015 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the realizeQNTF() function
"""

from __future__ import division, print_function

import numpy as np

from numpy.random import rand

from ._evalTF import evalTF
from ._partitionABCD import partitionABCD
from ._utils import _get_zpk, diagonal_indices

def realizeQNTF(ntf, form='FB', rot=False, bn=0.):
    """Convert a quadrature NTF into an ABCD matrix

    The basic idea is to equate the value of the loop filter at a set of
    points in the z-plane to the values of :math:`L_1 = 1-1/H` at those points.

    The order of the zeros is used when mapping the NTF onto the chosen
    topology.

    Supported structures are:

    * 'FB': Feedback
    * 'PFB': Parallel feedback
    * 'FF': Feedforward (bn is the coefficient of the aux DAC)
    * 'PFF': Parallel feedforward

    Not currently supported - feel free to send in a patch:

    * 'FBD': FB with delaying quantizer
    * 'FFD': FF with delaying quantizer

    **Parameters:**

    ntf : lti object or supported description
        The NTF to be converted.
    form : str, optional
        The form to be used. See above for the currently supported
        topologies. Defaults to ``'FB'``.
    rot : bool, optional
        Set ``rot`` to ``True`` to rotate the states to make as
        many coefficients as possible real.
    bn : float, optional
        Coefficient of the auxiliary DAC, to be specified for a
        'FF' form. Defaults to ``0.0``.

    **Returns:**

    ABCD : ndarray
        ABCD realization of the requested type.

    """
    #Code common to all forms
    ntf_z, ntf_p, _ = _get_zpk(ntf)
    ntf_z = np.sort_complex(ntf_z)
    ntf_p = np.sort_complex(ntf_p)
    order = ntf_p.shape[0]
    form = form.upper()

    A = np.diag(ntf_z)
    subdiag = diagonal_indices(A, -1)
    A[subdiag] = 1.
    if form == 'PFB' or form == 'PFF':
        # partition the NTF zeros into in-band and image-band
        for _ in range(2):
            # if the zeros are not correctly sorted, we sort them here
            fz = np.angle(ntf_z)/(2*np.pi)
            f0 = fz[0]
            inband_zeros = np.abs(fz - f0) < np.abs(fz + f0)
            ntf_z, inband_zeros = _move_zeros_to_the_end(ntf_z, inband_zeros)
        n_in = np.sum(inband_zeros)
        imageband_zeros = np.logical_not(inband_zeros)
        n_im = np.sum(imageband_zeros)
        if np.any(np.logical_not(imageband_zeros[n_in:])):
            # this should never happen.
            raise ValueError('Please put the image-band zeros at the end of' +
                             ' the ntf zeros')
        if n_im > 0:
            A[n_in, n_in-1] = 0.
    D = np.zeros(shape=(1, 2), dtype='complex128')

    # Find a set of points in the z-plane that are not close to zeros of H
    zSet = np.zeros((2*order, ), dtype='complex128')
    for i in range(2*order):
        z = 2*rand(1) - 1 + 1j*(2*rand(1) - 1)
        while np.any(np.abs(z - ntf_z) < 0.1) or np.any(np.abs(z - zSet)) < 0.1:
            z = 2*rand(1) - 1 + 1j*(2*rand(1) - 1)
        zSet[i] = z[0]
    # Evaluate L1 = 1-1/H at these points
    L1 = 1.0 - 1.0/evalTF(ntf, zSet)

    if form == 'FB':
        B = np.zeros(shape=(order, 2), dtype='complex128')
        C = np.hstack((np.zeros((1, order-1), dtype='complex128'),
                       np.atleast_2d(1)))
        # Compute F = C * inv(zI-A) at each z in zSet
        F = np.zeros((zSet.shape[0], order), dtype='complex128')
        I = np.eye(order)
        for i in range(zSet.shape[0]):
            F[i, :] = np.dot(C, np.linalg.inv(zSet[i]*I - A))
        B[:, 1] = np.linalg.lstsq(F, L1)[0]
        if rot == True:
            ABCD = np.vstack((np.hstack((A, B[:, 1].reshape((-1, 1)))),
                              np.hstack((C, np.atleast_2d(0)))))
            for i in range(order):
                phi = np.angle(ABCD[i, -1])
                ABCD[i, :] = ABCD[i, :]*np.exp(-1j*phi)
                ABCD[:, i] = ABCD[:, i]*np.exp(+1j*phi)
            A, B2, C, _ = partitionABCD(ABCD)
            B[:, 1] = np.squeeze(B2)
        B[0, 0] = np.abs(B[0, 1])
    elif form == 'PFB':
        B = np.zeros((order, 2), dtype='complex128')
        C = np.hstack((np.zeros((1, n_in-1)),
                       np.ones((1, 1)),
                       np.zeros((1, n_im-1)),
                       np.ones((1, 1))))
        # Compute F = C * inv(zI-A) at each z in zSet
        F = np.zeros((zSet.shape[0], order), dtype='complex128')
        I = np.eye(order)
        for i in range(zSet.shape[0]):
            F[i, :] = np.dot(C, np.linalg.inv(zSet[i]*I - A))
        B[:, 1] = np.linalg.lstsq(F, L1)[0]
        if rot == True:
            ABCD = np.vstack((np.hstack((A, B[:, 1].reshape((-1, 1)))),
                              np.hstack((C, np.atleast_2d(0)))))
            for i in range(order):
                phi = np.angle(ABCD[i, -1])
                ABCD[i, :] = ABCD[i, :]*np.exp(-1j*phi)
                ABCD[:, i] = ABCD[:, i]*np.exp(+1j*phi)
            A, B2, C, _ = partitionABCD(ABCD)
            B[:, 1] = np.squeeze(B2)
        B[0, 0] = np.abs(B[0, 1])
    elif form == 'FF':
        B0 = np.vstack((np.atleast_2d(1),
                        np.zeros((order-1, 1), dtype='complex128')))
        B = B0.copy()
        B[-1, 0] = bn
        # Compute F = inv(zI-A)*B at each z in zSet
        F = np.zeros((order, zSet.shape[0]), dtype='complex128')
        I = np.eye(order)
        for i in range(zSet.shape[0]):
            F[:, i] = np.squeeze(np.dot(np.linalg.inv(zSet[i]*I - A), B))
        L1 = L1.reshape((-1, 1))
        C = np.linalg.lstsq(F.T, L1)[0].T
        if rot == True:
            ABCD = np.vstack((np.hstack((A, B)),
                              np.hstack((C, np.atleast_2d(0)))))
            for i in range(1, order-1):
                phi = np.angle(ABCD[-1, i])
                ABCD[i, :] = ABCD[i, :]*np.exp(+1j*phi)
                ABCD[:, i] = ABCD[:, i]*np.exp(-1j*phi)
            A, B, C, _ = partitionABCD(ABCD)
        B = np.hstack((-B0, B))
    elif form == 'PFF':
        B0 = np.vstack((np.atleast_2d(1),
                        np.zeros((order-1, 1), dtype='complex128')))
        B = B0.copy()
        B[n_in] = 1.
        # Compute F = inv(zI-A)*B at each z in zSet
        F = np.zeros((order, zSet.shape[0]), dtype='complex128')
        I = np.eye(order)
        for i in range(zSet.shape[0]):
            F[:, i] = np.squeeze(np.dot(np.linalg.inv(zSet[i]*I - A), B))
        L1 = L1.reshape((-1, 1))
        C = np.linalg.lstsq(F.T, L1)[0].T
        C = C[:, ::-1]
        if rot == True:
            ABCD = np.vstack((np.hstack((A, B)),
                              np.hstack((C, np.zeros((1, 1))))))
            for i in range(1, order-1):
                phi = np.angle(ABCD[-1, i])
                ABCD[i, :] = ABCD[i, :]*np.exp(+1j*phi)
                ABCD[:, i] = ABCD[:, i]*np.exp(-1j*phi)
            A, B, C, _ = partitionABCD(ABCD)
        B = np.hstack((B0, B))
    else:
        raise ValueError('Sorry, form %s is not supported.' % form)
    ABCD = np.vstack((np.hstack((A, B)), np.hstack((C, D))))
    return ABCD

def _move_zeros_to_the_end(ntf_z, inband_zeros):
    """RealizeQNTF requires the imageband zeros to be put at the end
    of the list of zeros.

    This function ensures exactly that.
    """
    to_the_end = [i for i, iz in enumerate(inband_zeros) if not iz]
    for i, orig in enumerate(to_the_end):
        ntf_z = np.append(ntf_z, ntf_z[orig - i])
        ntf_z = np.delete(ntf_z, orig - i)
        inband_zeros = np.append(inband_zeros, inband_zeros[orig - i])
        inband_zeros = np.delete(inband_zeros, orig - i)
    return ntf_z, inband_zeros



from __future__ import division
import numpy as np
import deltasigma as ds
from scipy.signal import lti, ss2zpk, lfilter

def zpk_multiply(a, b):
    za, pa, ka = ds._utils._get_zpk(a)
    zb, pb, kb = ds._utils._get_zpk(b)
    pa = pa.tolist() if hasattr(pa, 'tolist') else pa
    pb = pb.tolist() if hasattr(pb, 'tolist') else pb
    za = za.tolist() if hasattr(za, 'tolist') else za
    zb = zb.tolist() if hasattr(zb, 'tolist') else zb
    return ds.cancelPZ((za+zb, pa+pb, ka*kb))

class testMultipleQ:
    def __init__(self):
        ABCD = [[1, 0, 0, 0, 1, -1, 0],
                [1, 1, 0, 0, 0, -2, 0],
                [0, 1, 1, 0, 0, 0, -1],
                [0, 0, 1, 1, 0, 0, -2],
                [0, 1, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 0, 0, 0]]
        self.ABCD = np.array(ABCD, dtype=np.float_)
        self.nlev = [9, 9]
        k = [1., 1.]
        self.ntfs, self.stfs = ds.calculateTF(self.ABCD, k)

    def testMultipleQ1(self):
        """Test function for DS simulation with nq>1 1/2"""
        v1n = zpk_multiply(self.ntfs[0, 0], ([2, -1], [1, 0, 0, 0, 0]))
        v2n = zpk_multiply(self.ntfs[1, 0], ([1, 1], [0, 0], 1))
        # compute v1n/v2n and check that it is equal to -1
        res = zpk_multiply(v1n, (ds._utils._get_zpk(v2n)[1], ds._utils._get_zpk(v2n)[0], 1./ds._utils._get_zpk(v2n)[2]))
        assert int(ds.pretty_lti(res)) == -1

    def testMultipleQ2(self):
        """Test function for DS simulation with nq>1 2/2"""
        # filtering and simulation
        filtM1 = [0., 0., 0., 2., -1.]
        filtM2 = [1., -2., 1.]
        ntf_eq = zpk_multiply(self.ntfs[1, 1], self.ntfs[1, 1])
        M = self.nlev[0] - 1
        osr = 64
        f0 = 0.
        f1, f2 = ds.ds_f1f2(OSR=64, f0=0., complex_flag=False)
        delta = 2
        Amp = ds.undbv(-3) # Test tone amplitude, relative to full-scale.
        f = 0.3 # will be adjusted to a bin
        N = 2**12
        f1_bin = np.round(f1*N)
        f2_bin = np.round(f2*N)
        fin = np.round(((1 - f)/2*f1 + (f + 1)/2*f2) * N)
        # input sine
        t = np.arange(0, N).reshape((1, -1))
        u = Amp*M*np.cos((2*np.pi/N)*fin*t)
        vx, _, xmax, y = ds.simulateDSM(u, self.ABCD, nlev=self.nlev)
        # separate output #1 and output #2
        v1 = vx[0, :]
        v2 = vx[1, :]
        # filter and combine
        vf = lfilter(filtM1, [1.], v1) + lfilter(filtM2, [1.], v2)
        # compute the spectra
        window = ds.ds_hann(N)
        NBW = 1.5/N
        spec0 = np.fft.fft(vf*window)/(M*N/2)/ds.undbv(-6)
        spec1 = np.fft.fft(v1*window)/(M*N/2)/ds.undbv(-6)
        spec2 = np.fft.fft(v1*window)/(M*N/2)/ds.undbv(-6)
        freq = np.linspace(0, 0.5, N/2 + 1)

        # smooth, calculate the theorethical response and the SNR for VF
        spec0_smoothed = ds.circ_smooth(np.abs(spec0)**2., 16)
        Snn0 = np.abs(ds.evalTF(ntf_eq, np.exp(2j*np.pi*freq)))**2 * 2/12*(delta/M)**2
        snr0 = ds.calculateSNR(spec0[f1_bin:f2_bin + 1], fin - f1_bin)

        # smooth, calculate the theorethical response and the SNR for V1
        spec1_smoothed = ds.circ_smooth(np.abs(spec1)**2., 16)
        Snn1 = np.abs(ds.evalTF(self.ntfs[0, 0], np.exp(2j*np.pi*freq)))**2 * 2/12*(delta/M)**2
        snr1 = ds.calculateSNR(spec1[f1_bin:f2_bin + 1], fin - f1_bin)

        assert snr0 > 40
        assert snr1 > 40
        assert snr0-snr1 > 40




# -*- coding: utf-8 -*-
# _simulateDSM_helper.pyxi
# Helper inline functions for the cython implementations of simulateDSM
#
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.
#
# This file is originally from `pydsm`. Little modifications have been
# performed - most prominently a bug in the ABCD matrix handling has been
# fixed. Many thanks to the original author.
#
# The original file is:
# Copyright (c) 2012, Sergio Callegari
# All rights reserved.
#
# The modifications are:
# Copyright (c) 2014, G. Venturini and the python-deltasigma contributors
#

cdef inline double dbl_sat(double x, double a, double b):
    return a if x <= a else b if x >= b else x

cdef inline void ds_quantize(int N, double* y, int y_stride, \
                             int* n, int n_stride, \
                             double* v, int v_stride):
    """Quantize a signal according to a given number of levels."""
    cdef int qi
    cdef double L
    for qi in range(N):
        if n[2*qi*n_stride] % 2 == 0:
            v[qi*v_stride] = 2*floor(0.5*y[qi*y_stride]) + 1
        else:
            v[qi*v_stride] = 2*floor(0.5*(y[qi*y_stride] + 1))
        L = n[2*qi*n_stride] - 1
        v[qi*v_stride] = dbl_sat(v[qi*v_stride], -L, L)

cdef inline void track_vabsmax(int N,\
                               double* vabsmax, int vabsmax_stride,\
                               double* x, int x_stride):
    cdef int i
    cdef double absx
    for i in range(N):
        absx = fabs(x[i*x_stride])
        if absx > vabsmax[i*vabsmax_stride]:
            vabsmax[i*vabsmax_stride] = absx

cdef inline double *dbldata(np.ndarray arr):
    return <double *>np.PyArray_DATA(arr)

cdef inline int *intdata(np.ndarray arr):
    return <int *>np.PyArray_DATA(arr)



# -*- coding: utf-8 -*-
# _simulateSNR.py
# Module providing the simulateSNR function
# Copyright 2013 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the simulateSNR() function
"""

from __future__ import division, print_function

import collections
from warnings import warn

import numpy as np
from numpy.fft import fft, fftshift

from ._calculateSNR import calculateSNR
from ._mapQtoR import mapQtoR
from ._simulateDSM import simulateDSM
from ._simulateQDSM import simulateQDSM
from ._utils import _get_zpk


def simulateSNR(arg1, osr, amp=None, f0=0, nlev=2, f=None, k=13,
                quadrature=False):
    """Determine the SNR for a delta-sigma modulator by using simulations.

    Simulate a delta-sigma modulator with sine wave inputs of various
    amplitudes and calculate the signal-to-noise ratio (SNR) in dB for each 
    input.

    Three alternative descriptions of the modulator can be used:

     * The modulator is described by a noise transfer function (NTF), provided
       as ``arg1`` and the number of quantizer levels (``nlev``).

     * Alternatively, the first argument to simulateSNR may be an ABCD matrix. 

     * Lastly, ``arg1`` may be a function taking the input signal as its
       sole argument.

    The band of interest is defined by the oversampling ratio (``osr``)
    and the center frequency (``f0``).

    The input signal is characterized by the ``amp`` vector and the ``f`` variable.
    A default value for ``amp`` is used if not supplied.

    ``f`` is the input frequency, normalized such that 1 -> fs;
    ``f`` is rounded to an FFT bin.

    Using sine waves located in FFT bins, the SNR is calculated as the ratio
    of the sine wave power to the power in all in-band bins other than those
    associated with the input tone. Due to spectral smearing, the input tone
    is not allowed to lie in bins 0 or 1. The length of the FFT is :math:`2^k`.

    If the NTF is complex, :func:`simulateQDSM` (which is slow, also available
    in a future release) is called.

    If ABCD is complex, :func:`simulateDSM` is used with the real equivalent
    of ABCD in order to speed up simulations.

    Future versions may accommodate STFs.

    **Parameters:**

    arg1 : scipy 'lti' object, or ndarray
        The first argument may be one of the various supported representations
        for an NTF or an ABCD matrix. Quadrature modulators are supported both
        with quadrature TFs or with quadrature ABCD matrices, the latter being
        the recommended description.

    osr : int
        The over-sampling ratio.

    amp : sequence, optional
        The amplitudes in dB, referred to the FS, for which the SNR is to be 
        evaluated. ``amp`` defaults to [-120 -110...-20 -15 -10 -9 -8 ... 0]dB,
        where 0 dB means a full-scale (peak value = nlev-1) sine wave.

    f0 : float, optional
        The center frequency. Normalized. Defaults to 0.

    nlev : int, optional
        Number of quantizer levels, defaults to 2.

    f : float, optional
        Test signal input frequency. Normalized. Rounded to an FFT bin.
        Defaults to:

    .. math::

        f = \\frac{1}{4\ \mathrm{OSR}}

    k : int, optional
        The number of samples used to compute the FFT is set by the integer `k`
        - default value 13 - through:

    .. math::

        N_{samples} = 2^k

    quadrature : boolean, optional
        Whether the delta sigma modulator is a quadrature modulator or not.
        Defaults to ``False``.

    **Returns:**

    snr : ndarray
        The SNR, from simulation.

    amp : ndarray
        The amplitudes corresponding to the SNR values.

    .. rubric:: Example:

    Compare the SNR vs input amplitude curve for a fifth-order modulator, as 
    determined by the describing function method (:func:`predictSNR`) with
    that determined by simulation (:func:`simulateSNR`).::

        import pylab as plt
        from deltasigma import *
        OSR = 32
        H = synthesizeNTF(5, OSR, 1)
        snr_pred, amp, _, _, _ = predictSNR(H,OSR)
        snr, amp = simulateSNR(H, OSR)
        plt.plot(amp, snr_pred, 'b', label='Predicted')
        plt.hold(True)
        plt.plot(amp, snr, 'go', label='Simulated')
        plt.grid(True)
        figureMagic([-100, 0], 10, None,
                    [0, 100], 10, None)
        plt.xlabel('Input Level, dB')
        plt.ylabel('SNR, dB')
        s = 'peak SNR = %4.1fdB\\n' % max(snr)
        plt.text(-65, 15, s, horizontalalignment='left')
        plt.legend(loc='best')

    .. plot::

        import pylab as plt
        from deltasigma import *
        OSR = 32
        H = synthesizeNTF(5, OSR, 1)
        snr_pred, amp, _, _, _ = predictSNR(H,OSR)
        snr, amp = simulateSNR(H, OSR)
        plt.plot(amp, snr_pred, 'b', label='Predicted')
        plt.hold(True)
        plt.plot(amp, snr, 'go', label='Simulated')
        plt.grid(True)
        figureMagic([-100, 0], 10, None,
                    [0, 100], 10, None)
        plt.xlabel('Input Level, dB')
        plt.ylabel('SNR, dB')
        s = 'peak SNR = %4.1fdB\\n' % max(snr)
        plt.text(-65, 15, s, horizontalalignment='left')
        plt.legend(loc='best')

    """
    # Look at arg1 and decide if the system is quadrature
    quadrature_ntf = False
    if callable(arg1):
        raise ValueError('There is no support for NTFs described through ' +
                         'a function.')
    elif not isinstance(arg1, np.ndarray):
        # NTF of LTI type or zpk tuple, or (num,den) etc...
        # in this case, if the modulator is in quadrature, 
        # we will have to call simulateQDSM
        for roots in _get_zpk(arg1)[:2]:
            if np.any(np.abs(np.imag(np.poly(roots))) > 0.0001):
                quadrature = True
                quadrature_ntf = True
    else: # ABCD matrix, no matter what, here we will get through the
        # simulation with simulateDSM()
        arg1 = np.real_if_close(arg1) # tiny imaginary parts due to rounding
        if not np.all(np.imag(arg1) == 0):
            quadrature = True

    if amp is None:
        amp = np.concatenate((np.arange(- 120, -20 + 1, 10),
                              np.array((-15,)),
                              np.arange(-10, 1)))
    elif not isinstance(amp, collections.Iterable):
        amp = np.array((amp, ))
    else:
        amp = np.asarray(amp)
    osr_mult = 2
    if f0 != 0 and not quadrature:
        osr_mult = 2*osr_mult
    if f is None or np.isnan(f):
        f = f0 + 0.5/(osr*osr_mult) # Halfway across the band
    M = nlev - 1
    if quadrature and not quadrature_ntf:
        # Modify arg1 (ABCD) and nlev so that simulateDSM can be used
        nlev = np.array([nlev, nlev])
        arg1 = mapQtoR(arg1)
    if abs(f - f0) > 1./(osr*osr_mult):
        warn('The input tone is out-of-band.')
    N = 2**k
    if N < 8*2*osr:
        warn('Increasing k to accommodate a large oversampling ratio.')
        k = np.array(np.ceil(np.log2(8*2*osr)), dtype=np.int64)
        N = 2**k
    F = np.round(f*N)
    if np.abs(F) <= 1:
        warn('Increasing k to accommodate a low input frequency.')
        # Want f*N >= 1
        k = np.ceil(np.log2(1./f))
        N = 2**k
        F = 2
    Ntransient = 100
    soft_start = 0.5*(1 - np.cos(2*np.pi/Ntransient * \
                                 np.arange(Ntransient/2)))
    if not quadrature:
        tone = M*np.sin(2*np.pi*F/N*np.arange(N + Ntransient))
        tone[:Ntransient/2] = tone[:Ntransient/2] * soft_start
    else:
        tone = M*np.exp(2j*np.pi*F/N * np.arange(N + Ntransient))
        tone[:Ntransient/2] = tone[:Ntransient/2] * soft_start
        if not quadrature_ntf:
            tone = tone.reshape((1, -1))
            tone = np.vstack((np.real(tone), np.imag(tone)))
    # create a Hann window
    window = 0.5*(1 - np.cos(2*np.pi*np.arange(N)/N))
    if f0 == 0:
        # Exclude DC and its adjacent bin
        inBandBins = int(N/2) + np.arange(3,
                                     np.round(N/osr_mult/osr) + 1,
                                     dtype=np.int32)
        F = F - 2
    else:
        f1 = np.round(N*(f0 - 1./osr_mult/osr))
        # Should exclude DC
        inBandBins = int(N/2) + np.arange(f1,
                                     np.round(N*(f0 + 1./osr_mult/osr)) + 1,
                                     dtype=np.int32)
        F = F - f1 + 1
    snr = np.zeros(amp.shape)
    for i, A in enumerate(np.power(10.0, amp/20)):
        if quadrature_ntf:
            v, _, _, _ = simulateQDSM(A*tone, arg1, nlev)
        else:
            v, _, _, _ = simulateDSM(A*tone, arg1, nlev)
            if quadrature:
                v = v[0, :] + 1j*v[1, :]
        hwfft = fftshift(fft(window*v[Ntransient:N + Ntransient]))
        snr[i] = calculateSNR(hwfft[inBandBins - 1], F)
    return snr, amp




from __future__ import division
from ._LCparam2tf import LCparam2tf

def LCoptparam2tf(x, param):
    """Convert optimization parameters to transfer functions."""
    n = param['n']
    param.update({'gw':np.ones(n-1)})
    if param['form'] == 'FB':
        param.update({'gu':np.concatenate((np.array([1]), np.zeros(n)))})
        param.update({'gv':copy.copy(x)})
        param.update({'gx':np.concatenate((np.zeros(n-1), np.array([1])))
        param.update({'rx':np.zeros(n)})
    elif param['form'] == 'FF':
        param.update({'gu':np.concatenate((np.array([1]), np.zeros(n-1),
                                           np.array([1])))})
        param.update({'gv':np.concatenate((np.array([1]), np.zeros(n))})
        param.update({'gx':copy.copy(x)
        param.update({'rx':np.zeros(n)})
    elif param['form'] == 'FFR':
        param.update({'gu':np.concatenate((np.array([1]), np.zeros(n-1),
                                           np.array([1])))})
        param.update({'gv':np.concatenate((np.array(x[0]), np.zeros(n)))})
        param.update({'gx':np.concatenate((np.zeros(n-1), np.array([1])))})
        param.update({'rx':np.concatenate((np.array([0]), x[1:n]))})
    elif param['form'] == 'GEN':
        param.update({'gv':np.atleast_1d(x[:n])})
        param.update({'rx':np.concatenate((np.array([0]), x[n:2*n-1]))})
    else:
        raise ValueError('form=%s is not supported.' % form)
    H, L0, ABCD = LCparam2tf(param)
    return H, L0, ABCD, param



# -*- coding: utf-8 -*-
# _calculateQTF.py
# Module providing the calculateQTF function
# Copyright 2013 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the calculateQTF() function
"""

from __future__ import division

import numpy as np

from scipy.signal import ss2tf, tf2zpk

from ._cancelPZ import cancelPZ
from ._partitionABCD import partitionABCD


def calculateQTF(ABCDr):
    """Calculate noise and signal transfer functions for a quadrature modulator

    **Parameters:**

    ABCDr : ndarray
        The ABCD matrix, in real form. You may call :func:`mapQtoR` to convert
        an imaginary (quadrature) ABCD matrix to a real one.

    **Returns:**

    ntf, stf, intf, istf : tuple of zpk tuples
        The quadrature noise and signal transfer functions.

    :raises RuntimeError: if the supplied ABCD matrix results in denominator mismatches.
    """
    A, B, C, D = partitionABCD(ABCDr, 4)

    #Construct an ABCD description of the closed-loop system
    # sys is a tuple in A, B, C, D form
    Acl = A + np.dot(B[:, 2:4], C)
    Bcl = B
    Ccl = C
    Dcl = np.hstack((D[:, 0:2], np.eye(2)))
    #sys = (A + np.dot(B[:, 2:4], C), B, C,
    #       np.hstack((D[:, 0:2], np.eye(2))))

    #Calculate the 2x4 matrix of transfer functions
    tfs = np.empty((2, 4), dtype=object)
    # Each tf is a tuple in num, den form
    # tf[i, j] corresponds to the TF from input j to output i
    for i in range(2):
        for j in range(4):
            tfs[i, j] = ss2tf(Acl, Bcl, Ccl[i, :], Dcl[i, :], input=j)

    #Reduce these to NTF, STF, INTF and ISTF
    if any(tfs[0, 2][1] != tfs[1, 3][1]):
        raise RuntimeError('TF Denominator mismatch. Location 1')

    ntf_x = (0.5 * (tfs[0, 2][0] + tfs[1, 3][0]), tfs[0, 2][1])
    intf_x = (0.5 * (tfs[0, 2][0] - tfs[1, 3][0]), tfs[0, 2][1])

    if any(tfs[0, 3][1] != tfs[1, 2][1]):
        raise RuntimeError('TF Denominator mismatch. Location 2')

    ntf_y = (0.5 * (tfs[1, 2][0] - tfs[0, 3][0]), tfs[1, 2][1])
    intf_y = (0.5 * (tfs[1, 2][0] + tfs[0, 3][0]), tfs[1, 2][1])

    if any(ntf_x[1] != ntf_y[1]):
        raise RuntimeError('TF Denominator mismatch. Location 3')
    if any(tfs[0, 0][1] != tfs[1, 1][1]):
        raise RuntimeError('TF Denominator mismatch. Location 4')

    stf_x = (0.5 * (tfs[0, 0][0] + tfs[1, 1][0]), tfs[0, 0][1])
    istf_x = (0.5 * (tfs[0, 0][0] - tfs[1, 1][0]), tfs[0, 0][1])

    if any(tfs[0, 1][1] != tfs[1, 0][1]):
        raise RuntimeError('TF Denominator mismatch. Location 5')

    stf_y = (0.5 * (tfs[1, 0][0] - tfs[0, 1][0]), tfs[1, 0][1])
    istf_y = (0.5 * (tfs[1, 0][0] + tfs[0, 1][0]), tfs[1, 0][1])

    if any(stf_x[1] != stf_y[1]):
        raise RuntimeError('TF Denominator mismatch. Location 6')

    # suppress warnings about complex TFs
    #warning('off')
    ntf = cancelPZ(tf2zpk(ntf_x[0] + 1j*ntf_y[0], ntf_x[1]))
    intf = cancelPZ(tf2zpk(intf_x[0] + 1j*intf_y[0], intf_x[1]))
    stf = cancelPZ(tf2zpk(stf_x[0] + 1j*stf_y[0], ntf_x[1]))
    istf = cancelPZ(tf2zpk(istf_x[0] + 1j*istf_y[0], intf_x[1]))
    #warning('on')

    return ntf, stf, intf, istf



# -*- coding: utf-8 -*-
# test_scaleABCD.py
# This module provides the tests for the scaleABCD function.
# Copyright 2014 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""This module provides the test class for the scaleABCD() function.
"""

from __future__ import division, print_function

import unittest
import numpy as np
import deltasigma as ds

class TestScaleABCD(unittest.TestCase):
    """Test class for scaleABCD()"""

    def setUp(self):
        order = 8
        osr = 32
        self.nlev = 2
        self.f0 = 0.125
        Hinf = 1.5
        form = 'CRFB'
        # Optimized zero placement
        ntf = ds.synthesizeNTF(order, osr, 2, Hinf, self.f0)
        a, g, b, c = ds.realizeNTF(ntf, form)
        # we pass b = b[0]
        # if you use a single feed-in for the input, b may be scalar too
        ABCD = ds.stuffABCD(a, g, b[0], c, form)
        # now we correct b for the assert
        # Use a single feed-in for the input
        b = np.concatenate((np.atleast_1d(b[0]),
                            np.zeros((max(b.shape) - 1,))))
        self.ABCD0 = ABCD.copy()

        # References
        self.Sdiag_ref = np.array([71.9580, 51.9359, 8.2133, 6.5398,
                                   1.9446, 1.2070, 0.4223, 0.3040])
        self.umax_ref = 0.8667
        ABCD_ref1 = np.array([[1., -0.7320, 0, 0, 0, 0],
                              [0.7218, 0.4717, 0, 0, 0, 0],
                              [0, 0.1581, 1., -0.7357, 0, 0],
                              [0, 0.1259, 0.7962, 0.4142, 0, 0],
                              [0, 0, 0, 0.2973, 1., -0.9437],
                              [0, 0, 0, 0.1846, 0.6207, 0.4142],
                              [0, 0, 0, 0, 0, 0.3499],
                              [0, 0, 0, 0, 0, 0.2518],
                              [0, 0, 0, 0, 0, 0]])
        ABCD_ref2 = np.array([[0, 0, 0.0858, -0.0858],
                              [0, 0, 0.0619, 0.0428],
                              [0, 0, 0, 0.0642],
                              [0, 0, 0, 0.1835],
                              [0, 0, 0, 0.2447],
                              [0, 0, 0, 0.0581],
                              [1., -0.8971, 0, -0.0076],
                              [0.7197, 0.3543, 0, -0.1746],
                              [0, 3.29, 0, 0]])
        self.ABCD_ref = np.hstack((ABCD_ref1, ABCD_ref2))

    def test_scaleABCD(self):
        """Test function for scaleABCD()"""
        ABCD, umax, S = ds.scaleABCD(self.ABCD0, self.nlev, self.f0)
        # mapping the NTF to states, there is not a perfect match between
        # the original code and the scipy version. -> rtol approx 20%
        #if not np.allclose(ABCD, ABCD_ref, atol=1e-2, rtol=3e-1):
        #    aerr = ABCD_ref-ABCD
        #    rerr = 2*(ABCD_ref-ABCD)/(ABCD_ref+ABCD)
        #    print(repr(ABCD_ref))
        #    print(repr(ABCD))
        #    print(aerr)
        #    print(rerr)
        # this is a rather high relative error. We get it on Travis-CI
        # Probably related to the libs used?
        self.assertTrue(np.allclose(ABCD, self.ABCD_ref, atol=1e-2, rtol=30e-2))
        self.assertTrue(np.allclose(umax, self.umax_ref, atol=1e-4, rtol=1e-3))
        self.assertTrue(np.allclose(np.diag(S), self.Sdiag_ref, atol=1e-2, rtol=25e-1))

    def test_scaleABCD_Q(self):
        """Test function for scaleABCD() with Imag ABCD"""
        ABCD_IN = np.array([[0.888+0.4598j, 0.+0.j,  0.+0.j, 0.+0.j, 0.0516+0.j,
                             0.0516+0.j],
                            [0.6073+0.7945j, 0.9530+0.3028j, 0.+0.j, 0.+0.j,
                             0.+0.j, 0.3230+0.j],
                            [0.+0.j, 0.9459+0.3245j, 0.9239+0.3827j, 0.+0.j,
                             0.+0.j, 0.9162+0.j],
                            [0.+0.j, 0.+0.j, 0.+0.j, 0.9239-0.3827j, 0.+0.j,
                             0.1778+0.j],
                            [0.+0.j, 0.+0.j, -0.9935-0.1138j, -0.7405-0.672j,
                             0.+0.j, 0.+0.j]])

        ABCD, umax, S = ds.scaleABCD(ABCD_IN, nlev=9, f=1./16)

        #references here because I prefer to have 1 file
        umax_ref = 8
        S_ref = np.array([[0.3586, 0., 0., 0.],
                          [0., 0.1207, 0., 0.],
                          [0., 0., 0.0805, 0.],
                          [0., 0., 0., 0.3671]])
        ABCD_ref = np.array([[0.888037198535288 + 0.459771610712905j,
                              0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j,
                              0.0185052748572140 + 0.0j,
                              0.0185052748572140 + 4.35415940995948e-18j],
                             [0.204472470161495 + 0.267520232240738j,
                              0.953044748762363 + 0.302829501298050j,
                              0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j,
                              0.0389934281708177 + 0.0j],
                             [0.0 + 0.0j,
                              0.630626110921381 + 0.216330153871036j,
                              0.923879532511340 + 0.382683432365112j,
                              0.0 + 0.0j, 0.0 + 0.0j,
                              0.0737451799430360 - 1.22879569706001e-17j],
                             [0.0 + 0.0j, 0.0 + 0.0j, 0.0 + 0.0j,
                              0.923879532511287 - 0.382683432365090j,
                              0.0 + 0.0j,
                              0.0652579452814600 + 1.01879713420556e-17j],
                             [0.0 + 0.0j, 0.0 + 0.0j, -12.3424213726332 -
                              1.41434395098423j, -2.01750453626270 -
                              1.83077728688503j, 0.0 + 0.0j, 0.0 + 0.0j]])

        assert np.allclose(ABCD, ABCD_ref, atol=1e-2, rtol=1e-1)
        assert np.allclose(S, S_ref, atol=1e-3, rtol=1e-2)
        assert umax == umax_ref




# -*- coding: utf-8 -*-
# _simulateQDSM_core.py
# Module providing the simulateQDSM function
# Copyright 2015 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the core of the simulateQDSM() function
"""

from __future__ import division, print_function

import numpy as np

from ._ds_quantize import ds_quantize

def simulateQDSM_core(u, A, B, C, D1, order, nlev, nq, x0):
    N = u.shape[1]
    v = np.zeros(shape=(nq, N), dtype='complex128')
    y = np.zeros(shape=(nq, N), dtype='complex128')
    # Need to store the state information
    xn = np.zeros(shape=(order, N), dtype='complex128')
    # Need to keep track of the state maxima
    xmax = abs(x0.copy())
    for i in range(N):
        y[:, i] = np.dot(C, x0) + np.dot(D1, u[:, i].reshape((-1, 1)))
        v[:, i] = ds_qquantize(y[:, i], nlev)
        x0 = np.dot(A, x0) + np.dot(B, np.vstack((u[:, i], v[:, i])))
        # Save the next state
        xn[:, i] = np.squeeze(x0)
        # Keep track of the state maxima
        xmax = np.max((np.abs(x0), xmax), axis=0)
    return v, xn, xmax, y

def ds_qquantize(y, n):
    """Quadrature quantization
    """
    if np.isreal(n):
        v = ds_quantize(np.real(y), n) + 1j*ds_quantize(np.imag(y), n)
    else:
        ytmp = np.concatenate((np.real(y) + np.imag(y),
                               np.real(y) - np.imag(y)))
        v = np.dot(ds_quantize(ytmp, np.abs(n)),
                   np.array([[1 + 1j], [1 - 1j]]))/2.
    return v




# -*- coding: utf-8 -*-
# _simulateQDSM.py
# Module providing the simulateQDSM function
# Copyright 2015 Giuseppe Venturini
# This file is part of python-deltasigma.
#
# python-deltasigma is a 1:1 Python replacement of Richard Schreier's
# MATLAB delta sigma toolbox (aka "delsigma"), upon which it is heavily based.
# The delta sigma toolbox is (c) 2009, Richard Schreier.
#
# python-deltasigma is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LICENSE file for the licensing terms.

"""Module providing the simulateQDSM() function
"""

from __future__ import division, print_function

import copy

import numpy as np

import scipy

from scipy.linalg import lstsq
from scipy.signal import freqz, tf2zpk

from ._config import _debug, setup_args
from ._ds_quantize import ds_quantize
from ._evalTF import evalTF
from ._partitionABCD import partitionABCD
from ._utils import carray, diagonal_indices, _is_zpk, _is_A_B_C_D, _is_num_den

# try to import and, if necessary, compile the Cython-optimized
# core of the simulation code`
try:
    import pyximport
    pyximport.install(setup_args=setup_args, inplace=True)
    from ._simulateQDSM_core import simulateQDSM_core
except ImportError as e:
    if _debug:
        print(str(e))
    # we'll just fall back to the Python version
    pass


def simulateQDSM(u, arg2, nlev=2, x0=None):
    """Simulate a quadrature Delta-Sigma modulator

    This function computes the output of a quadrature delta-sigma modulator
    corresponding to an input :math:`u`, with a description of the modulator, an
    initial state :math:`x_0` (default all zeros) and a quantizer whose number
    of levels is specified by :math:`n_{lev}`.

    For multiple quantizers, make :math:`n_{lev}` a 1D vector, for complex
    quantization to a diamond lattice, multiply :math:`n_{lev}` by :math:`j`.

    Regarding the description of the modulator, it may be provided through an
    ABCD matrix.

    In this case, the shapes of the input parameters are:

    * ``u.shape = (nu, N)``,
    * ``nlev.shape = (nqi,)``,
    * ``ABCD.shape = (order+nq, order+nq+nu)``.

    Alternatively, the modulator may be described by a supported TF
    representation, in particular it is recommended to use a zpk object. In this
    case, the STF is assumed to be 1.

    **Parameters:**

    u : ndarray
        The input signal to the modulator.
    arg2 : ndarray or a supported LTI representation
        A description of the modulator to simulate.
        An ndarray instance is interpreted as an ABCD description. Equivalently,
        the ABCD matrix may be supplied in ``(A, B, C, D)`` tuple form. All
        other supported modulator specifications result in a conversion to a zpk
        representation.
    nlev : int or sequence-like, optional
        The number of levels in the quantizer. If set to a sequence, each of the
        elements is assumed to be the number of levels associated with a
        quantizer. Defaults to ``2``.
    x0 : float or sequence-like, optional
        The initial states of the modulator. If it is set to a float, all states
        are assumed to have the same value, ``x0``. If it is set to a
        sequence-like objct (list, tuple, 1D ndarray and similar), each entry is
        assumed to be the value of one of the modulator states, in ascending
        order. Defaults to ``0``.

    **Returns:**

    v : ndarray
        The quantizer output.
    xn : ndarray
        The modulator states.
    xmax : nedarray
        The maximum value that each state reached during simulation.
    y : ndarray
        The quantizer input (ie the modulator output).
    """

    if len(u.shape) == 1:
        u = u.reshape((1, -1))
    nu = u.shape[0]
    if hasattr(nlev, '__len__'):
        nlev = np.atleast_1d(nlev)
        nq = max(nlev.shape)
    else:
        nq = 1
    if isinstance(arg2, scipy.signal.lti):
        k = arg2.k
        zeros = np.asarray(arg2.z)
        poles = np.asarray(arg2.p)
        form = 2
        order = max(zeros.shape)
    elif _is_zpk(arg2):
        zeros, poles, k = copy.deepcopy(arg2)
        zeros = np.asarray(zeros)
        poles = np.asarray(poles)
        form = 2
        order = max(zeros.shape)
    elif isinstance(arg2, np.ndarray): # ABCD
        if arg2.shape[1] > 2 and arg2.shape[1] == nu + arg2.shape[0]:
            # ABCD dimesions OK
            form = 1
            ABCD = arg2
            order = ABCD.shape[0] - nq
        else:
            raise ValueError('The ABCD argument does not have proper ' +
                             'dimensions.')
    elif _is_A_B_C_D(arg2):
        ABCD = np.vstack((np.hstack((np.atleast_2d(arg2[0]),
                                     np.atleast_2d(arg2[1]))),
                          np.hstack((np.atleast_2d(arg2[2]),
                                     np.atleast_2d(arg2[3])))))
        form = 1
        order = ABCD.shape[0] - nq
    elif _is_num_den(arg2):
        zeros, poles, k = tf2zpk(*arg2)
        form = 2
        order = max(zeros.shape)
    else:
        raise TypeError('The second argument is neither an ABCD matrix nor ' +
                        'an NTF.')

    if x0 is None:
        x0 = np.zeros(shape=(order, 1), dtype='complex128')
    else:
        x0 = carray(x0)
        x0 = np.atleast_2d(x0).astype('complex128')
    if form == 1:
        A, B, C, D = partitionABCD(ABCD, nq + nu)
        A = A.astype('complex128')
        B = B.astype('complex128')
        C = C.astype('complex128')
        D = D.astype('complex128')
        D1 = D[:, :nu].reshape((-1, nu))
    else:
        # Create a FF realization of 1-1/H.
        # Note that MATLAB's zp2ss and canon functions don't work for complex
        # TFs.
        A = np.zeros(shape=(order, order), dtype='complex128')
        B2 = np.vstack((np.atleast_2d(1), np.zeros(shape=(order-1, 1),
                                                   dtype='complex128')))
        diag = diagonal_indices(A, 0)
        A[diag] = zeros
        subdiag = diagonal_indices(A, -1)
        A[subdiag] = 1.
        # Compute C st C*inv(zI-A)*B = 1-1/H(z);
        w = 2*np.pi*np.random.rand(2*order)
        desired = 1 - 1.0/evalTF((zeros, poles, k), np.exp(1j*w))
        desired.reshape((1, -1))
        # suppress warnings about complex TFs ???
        sysresp = np.zeros((order, w.shape[0]), dtype='complex128')
        for i in range(order):
            Ctemp = np.zeros((1, order))
            Ctemp[0, i] = 1
            sys = (A, B2, Ctemp, np.zeros((1, 1)))
            n, d = scipy.signal.ss2tf(*sys)
            sysresp[i, :] = freqz(n[0, :], d, w)[1]
        C = lstsq(sysresp.T, desired.T)[0].reshape((1, -1))
        # !!!! Assume stf=1
        B1 = -B2
        B = np.hstack((B1, B2))
        D1 = np.ones((1, 1), dtype='complex128')
    v, xn, xmax, y = simulateQDSM_core(u, A, B, C, D1, order, nlev, nq, x0)
    return v.squeeze(), xn.squeeze(), xmax, y.squeeze()

